import abc
from flask import Flask, request, send_file
from flask.ext.restful import abort, Api, Resource
from flask.ext.restful.utils import cors


class LandmarkerIOAdapter(object):
    r"""
    Abstract definition of an adapter that can be passed to app_for_adapter in
    order to generate a legal Flask implementation of landmarker.io's REST API.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def mesh_ids(self):
        pass

    @abc.abstractmethod
    def mesh_json(self, mesh_id):
        pass

    @abc.abstractmethod
    def textured_mesh_ids(self):
        pass

    @abc.abstractmethod
    def texture_file(self, mesh_id):
        pass
    @abc.abstractmethod
    def all_landmarks(self):
        pass

    @abc.abstractmethod
    def landmark_ids(self, mesh_id):
        pass

    @abc.abstractmethod
    def landmark_json(self, mesh_id, lm_id):
        pass

    @abc.abstractmethod
    def save_landmark_json(self, mesh_id, lm_id, lm_json):
        pass

    @abc.abstractmethod
    def templates(self):
        pass

    @abc.abstractmethod
    def template_json(self, lm_id):
        pass


def app_for_adapter(adapter, gzip=False, dev=False):
    r"""
    Generate a Flask App that will serve meshes landmarks and templates to
    landmarker.io

    Parameters
    ----------

    adapter: :class:`LandmarkerIOAdapter`
        Concrete implementation of the LandmarkerIOAdapter. Will be queried for
        all data to pass to landmarker.io.

    gzip: Boolean, optional
        If True, responses will be gzipped before being sent to the client.
        Higher workload for the server, smaller payload to the client.

        Default: False

    """
    app = Flask(__name__)

    if gzip:
        from flask.ext.compress import Compress
        Compress(app)

    api = Api(app)
    origin = 'http://www.landmarker.io'
    if dev:
        # in development mode, accept CORS from anyone
        origin = '*'
    api.decorators = [cors.crossdomain(origin=origin,
                                       headers=['Content-Type'])]

    class Mesh(Resource):

        def get(self, mesh_id):
            try:
                return adapter.mesh_json(mesh_id)
            except:
                abort(404, message="{} is not an available model".format(mesh_id))

    class MeshList(Resource):

        def get(self):
            return adapter.mesh_ids()

    class Texture(Resource):

        def get(self, mesh_id):
            try:
                return send_file(adapter.texture_file(mesh_id),
                                 mimetype='image/jpeg')
            except:
                abort(404, message="{} is not a textured mesh".format(mesh_id))

    class TextureList(Resource):

        def get(self):
            return adapter.textured_mesh_ids()

    class Landmark(Resource):

        def get(self, mesh_id, lm_id):
            try:
                return adapter.landmark_json(mesh_id, lm_id)
            except:
                abort(404, message="{}:{} does not exist".format(mesh_id, lm_id))

        def put(self, mesh_id, lm_id):
            return adapter.save_landmark_json(mesh_id, lm_id, request.json)

    class LandmarkList(Resource):

        def get(self):
            print 'asked for list'
            return adapter.all_landmarks()

    class LandmarkListForId(Resource):

        def get(self, mesh_id):
            return adapter.landmark_ids(mesh_id)

    class Template(Resource):

        def get(self, lm_id):
            try:
                return adapter.template_json(lm_id)
            except:
                abort(404, message="{} template not exist".format(lm_id))

    class TemplateList(Resource):

        def get(self):
            return adapter.templates()

    api_endpoint = '/api/v1/'

    api.add_resource(MeshList, api_endpoint + 'meshes')
    api.add_resource(Mesh, api_endpoint + 'meshes/<string:mesh_id>')

    api.add_resource(TextureList, api_endpoint + 'textures')
    api.add_resource(Texture, api_endpoint + 'textures/<string:mesh_id>')

    api.add_resource(LandmarkList, api_endpoint + 'landmarks')
    api.add_resource(LandmarkListForId, api_endpoint +
                     'landmarks/<string:mesh_id>')
    api.add_resource(Landmark, api_endpoint +
                     'landmarks/<string:mesh_id>/<string:lm_id>')

    api.add_resource(TemplateList, api_endpoint + 'templates')
    api.add_resource(Template, api_endpoint + 'templates/<string:lm_id>')

    return app



from copy import deepcopy
from collections import defaultdict
import glob
import json
import os
import os.path as p
import StringIO

import menpo.io as mio
from menpo.shape.mesh import TexturedTriMesh

from .utils import load_template
from .api import LandmarkerIOAdapter


def as_jpg_file(image):
    p = image.as_PILImage()
    output = StringIO.StringIO()
    p.save(output, format='jpeg')
    output.seek(0)
    return output


class MenpoAdapter(LandmarkerIOAdapter):

    def __init__(self, model_dir, landmark_dir, template_dir):
        self.model_dir = model_dir
        self.landmark_dir = landmark_dir
        self.template_dir = template_dir
        print ('models:    {}'.format(model_dir))
        print ('landmarks: {}'.format(landmark_dir))
        print ('templates: {}'.format(template_dir))

    def landmark_fp(self, model_id, lm_id):
        return p.join(self.landmark_dir, model_id, lm_id + '.json')

    def landmark_paths(self, mesh_id=None):
        if mesh_id is None:
            mesh_id = '*'
        g = glob.glob(p.join(self.landmark_dir, mesh_id, '*'))
        return filter(lambda f: p.isfile(f) and
                                p.splitext(f)[-1] == '.json', g)

    def mesh_paths(self):
        return mio.mesh_paths(p.join(self.model_dir, '*'))

    def texture_paths(self):
        return mio.image_paths(p.join(self.model_dir, '*'))

    def mesh_ids(self):
        return [p.splitext(p.split(m)[1])[0] for m in self.mesh_paths()]

    def mesh_json(self, mesh_id):
        mesh_glob = p.join(self.model_dir, mesh_id + '.*')
        return list(mio.import_meshes(mesh_glob))[0].tojson()

    def textured_mesh_ids(self):
        return [p.splitext(p.split(t)[1])[0] for t in self.texture_paths()]

    def texture_file(self, mesh_id):
        img_glob = p.join(self.model_dir, mesh_id + '.*')
        return as_jpg_file(list(mio.import_images(img_glob))[0])

    def all_landmarks(self):
        landmark_files = self.landmark_paths()
        mapping = defaultdict(list)
        for lm_path in landmark_files:
            dir_path, filename = p.split(lm_path)
            lm_set = p.splitext(filename)[0]
            lm_id = p.split(dir_path)[1]
            mapping[lm_id].append(lm_set)
        return mapping

    def landmark_ids(self, mesh_id):
        landmark_files = self.landmark_paths(mesh_id=mesh_id)
        return [p.splitext(p.split(f)[-1])[0] for f in landmark_files]

    def landmark_json(self, mesh_id, lm_id):
        fp = self.landmark_fp(mesh_id, lm_id)
        if not p.isfile(fp):
            raise IOError
        with open(fp, 'rb') as f:
            lm = json.load(f)
            return lm

    def save_landmark_json(self, mesh_id, lm_id, lm_json):
        subject_dir = p.join(self.landmark_dir, mesh_id)
        if not p.isdir(subject_dir):
            os.mkdir(subject_dir)
        fp = self.landmark_fp(mesh_id, lm_id)
        with open(fp, 'wb') as f:
            json.dump(lm_json, f, sort_keys=True, indent=4,
                      separators=(',', ': '))

    def templates(self):
        template_paths = glob.glob(p.join(self.template_dir, '*.lmt'))
        print self.template_dir
        print template_paths
        return [p.splitext(p.split(t)[-1])[0] for t in template_paths]

    def template_json(self, lm_id):
        fp = p.join(self.template_dir, lm_id + '.lmt')
        return load_template(fp)


class CachingMenpoAdapter(MenpoAdapter):

    def __init__(self, model_dir, landmark_dir, template_dir):
        MenpoAdapter.__init__(self, model_dir, landmark_dir, template_dir)
        print('Caching meshes and textures...')
        self.meshes = {}
        self.textures = {}
        for mesh in mio.import_meshes(p.join(self.model_dir, '*')):
            mesh_id = mesh.ioinfo.filename
            self.meshes[mesh_id] = mesh.tojson()
            if isinstance(mesh, TexturedTriMesh):
                self.textures[mesh_id] = as_jpg_file(mesh.texture)
        print(' - {} meshes imported.'.format(len(self.meshes)))
        print(' - {} meshes are textured.'.format(len(self.textures)))

    def mesh_ids(self):
        return list(self.meshes)

    def mesh_json(self, mesh_id):
        return self.meshes[mesh_id]

    def textured_mesh_ids(self):
        return list(self.textures)

    def texture_file(self, mesh_id):
        return deepcopy(self.textures[mesh_id])



import itertools
from collections import namedtuple

Group = namedtuple('Group', ['label', 'n', 'index'])


def parse_group(group):
    x = group.split('\n')
    label, n_str = x[0].split(' ')
    n = int(n_str)
    index_str = x[1:]
    if len(index_str) == 0:
        return Group(label, n, [])
    index = [[int(j) for j in i.split(' ')] for i in index_str]
    indexes = set(itertools.chain.from_iterable(index))
    if min(indexes) < 0 or max(indexes) > n:
        raise ValueError("invalid connectivity")
    return Group(label, n, index)


def group_to_json(group):
    group_json = {}
    lms = [{'point': None}] * group.n
    group_json["landmarks"] = lms
    group_json["connectivity"] = group.index
    return group_json


def groups_to_json(groups):
    lm_json = {'version': 1, 'groups': {}}
    for g in groups:
        lm_json['groups'][g.label] = group_to_json(g)
    return lm_json


def load_template(path):
    with open(path, 'rb') as f:
        ta = f.read().strip().split('\n\n')
    print ta
    groups = [parse_group(g) for g in ta]
    return groups_to_json(groups)



from setuptools import setup, find_packages
import os.path as p

setup(name='landmarkerio-server',
      version='0.0.2',
      description='Menpo-based server for www.landmarker.io',
      author='James Booth',
      author_email='james.booth08@imperial.ac.uk',
      url='https://github.com/menpo/landmarkerio-server/',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
      ],
      packages=['landmarkerioserver'],
      setup_requires=['menpo>=0.2.5'
                      #'Flask>=0.10.1',
                      #'Flask-Compress>=1.0.0',
                      #'Flask-RESTful>=0.2.11'
                      ],
      scripts=[p.join('landmarkerioserver', 'landmarkerio')]
      )
